"use strict";
/*!
  Copyright 2018 Ron Buckton (rbuckton@chronicles.org)

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
 */
/** @module "iterable-query/fn" */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __asyncValues = (this && this.__asyncValues) || function (o) {
    if (!Symbol.asyncIterator) throw new TypeError("Symbol.asyncIterator is not defined.");
    var m = o[Symbol.asyncIterator], i;
    return m ? m.call(o) : (o = typeof __values === "function" ? __values(o) : o[Symbol.iterator](), i = {}, verb("next"), verb("throw"), verb("return"), i[Symbol.asyncIterator] = function () { return this; }, i);
    function verb(n) { i[n] = o[n] && function (v) { return new Promise(function (resolve, reject) { v = o[n](v), settle(resolve, reject, v.done, v.value); }); }; }
    function settle(resolve, reject, d, v) { Promise.resolve(v).then(function(v) { resolve({ value: v, done: d }); }, reject); }
};
Object.defineProperty(exports, "__esModule", { value: true });
const internal_1 = require("../internal");
const lastAsync_1 = require("./lastAsync");
/**
 * Finds the value at the provided offset. A negative offset starts from the
 * last element.
 *
 * @param source An [[AsyncQueryable]] object.
 * @param offset An offset.
 * @category Scalar
 */
function elementAtAsync(source, offset) {
    return __awaiter(this, void 0, void 0, function* () {
        var e_1, _a, e_2, _b;
        internal_1.assert.mustBeAsyncQueryable(source, "source");
        internal_1.assert.mustBeInteger(offset, "offset");
        if (offset === -1) {
            return yield lastAsync_1.lastAsync(source);
        }
        if (offset < 0) {
            offset = Math.abs(offset);
            const array = [];
            try {
                for (var _c = __asyncValues(internal_1.ToPossiblyAsyncIterable(source)), _d; _d = yield _c.next(), !_d.done;) {
                    const element = _d.value;
                    if (array.length >= offset) {
                        array.shift();
                    }
                    array.push(element);
                }
            }
            catch (e_1_1) { e_1 = { error: e_1_1 }; }
            finally {
                try {
                    if (_d && !_d.done && (_a = _c.return)) yield _a.call(_c);
                }
                finally { if (e_1) throw e_1.error; }
            }
            return array.length - offset >= 0 ? array[array.length - offset] : undefined;
        }
        try {
            for (var _e = __asyncValues(internal_1.ToPossiblyAsyncIterable(source)), _f; _f = yield _e.next(), !_f.done;) {
                const element = _f.value;
                if (offset === 0) {
                    return element;
                }
                offset--;
            }
        }
        catch (e_2_1) { e_2 = { error: e_2_1 }; }
        finally {
            try {
                if (_f && !_f.done && (_b = _e.return)) yield _b.call(_e);
            }
            finally { if (e_2) throw e_2.error; }
        }
        return undefined;
    });
}
exports.elementAtAsync = elementAtAsync;
