"use strict";
/*!
  Copyright 2018 Ron Buckton (rbuckton@chronicles.org)

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
 */
/** @module "iterable-query/fn" */
var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
Object.defineProperty(exports, "__esModule", { value: true });
const internal_1 = require("../internal");
const union_1 = require("./union");
const map_1 = require("./map");
const lookup_1 = require("../lookup");
const defaultIfEmpty_1 = require("./defaultIfEmpty");
const common_1 = require("./common");
/**
 * Creates an [[Iterable]] for the correlated elements between an outer [[Queryable]] object and an inner
 * [[Queryable]] object.
 *
 * @param outer A [[Queryable]] object.
 * @param inner A [[Queryable]] object.
 * @param outerKeySelector A callback used to select the key for an element in `outer`.
 * @param innerKeySelector A callback used to select the key for an element in `inner`.
 * @param resultSelector A callback used to select the result for the correlated elements.
 * @category Join
 */
function fullJoin(outer, inner, outerKeySelector, innerKeySelector, resultSelector) {
    internal_1.assert.mustBeQueryable(outer, "outer");
    internal_1.assert.mustBeQueryable(inner, "inner");
    internal_1.assert.mustBeFunction(outerKeySelector, "outerKeySelector");
    internal_1.assert.mustBeFunction(innerKeySelector, "innerKeySelector");
    internal_1.assert.mustBeFunction(resultSelector, "resultSelector");
    return new FullJoinIterable(internal_1.ToIterable(outer), internal_1.ToIterable(inner), outerKeySelector, innerKeySelector, resultSelector);
}
exports.fullJoin = fullJoin;
let FullJoinIterable = class FullJoinIterable {
    constructor(outer, inner, outerKeySelector, innerKeySelector, resultSelector) {
        this._outer = outer;
        this._inner = inner;
        this._outerKeySelector = outerKeySelector;
        this._innerKeySelector = innerKeySelector;
        this._resultSelector = resultSelector;
    }
    *[Symbol.iterator]() {
        const resultSelector = this._resultSelector;
        const outerLookup = new lookup_1.Lookup(internal_1.CreateGroupings(this._outer, this._outerKeySelector, common_1.identity));
        const innerLookup = new lookup_1.Lookup(internal_1.CreateGroupings(this._inner, this._innerKeySelector, common_1.identity));
        const keys = union_1.union(map_1.map(outerLookup, internal_1.SelectGroupingKey), map_1.map(innerLookup, internal_1.SelectGroupingKey));
        for (const key of keys) {
            const outer = defaultIfEmpty_1.defaultIfEmpty(outerLookup.get(key), undefined);
            const inner = defaultIfEmpty_1.defaultIfEmpty(innerLookup.get(key), undefined);
            for (const outerElement of outer) {
                for (const innerElement of inner) {
                    yield resultSelector(outerElement, innerElement);
                }
            }
        }
    }
};
FullJoinIterable = __decorate([
    internal_1.ToStringTag("FullJoinIterable")
], FullJoinIterable);
