"use strict";
/*!
  Copyright 2018 Ron Buckton (rbuckton@chronicles.org)

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
 */
/** @module "iterable-query/fn" */
var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
var __await = (this && this.__await) || function (v) { return this instanceof __await ? (this.v = v, this) : new __await(v); }
var __asyncGenerator = (this && this.__asyncGenerator) || function (thisArg, _arguments, generator) {
    if (!Symbol.asyncIterator) throw new TypeError("Symbol.asyncIterator is not defined.");
    var g = generator.apply(thisArg, _arguments || []), i, q = [];
    return i = {}, verb("next"), verb("throw"), verb("return"), i[Symbol.asyncIterator] = function () { return this; }, i;
    function verb(n) { if (g[n]) i[n] = function (v) { return new Promise(function (a, b) { q.push([n, v, a, b]) > 1 || resume(n, v); }); }; }
    function resume(n, v) { try { step(g[n](v)); } catch (e) { settle(q[0][3], e); } }
    function step(r) { r.value instanceof __await ? Promise.resolve(r.value.v).then(fulfill, reject) : settle(q[0][2], r); }
    function fulfill(value) { resume("next", value); }
    function reject(value) { resume("throw", value); }
    function settle(f, v) { if (f(v), q.shift(), q.length) resume(q[0][0], q[0][1]); }
};
Object.defineProperty(exports, "__esModule", { value: true });
const internal_1 = require("../internal");
const lookup_1 = require("../lookup");
const union_1 = require("../fn/union");
const map_1 = require("../fn/map");
const defaultIfEmpty_1 = require("../fn/defaultIfEmpty");
const common_1 = require("./common");
/**
 * Creates an [[AsyncIterable]] for the correlated elements between an outer [[AsyncQueryable]] object and an inner
 * [[AsyncQueryable]] object.
 *
 * @param outer An [[AsyncQueryable]] object.
 * @param inner An [[AsyncQueryable]] object.
 * @param outerKeySelector A callback used to select the key for an element in `outer`.
 * @param innerKeySelector A callback used to select the key for an element in `inner`.
 * @param resultSelector A callback used to select the result for the correlated elements.
 * @category Join
 */
function fullJoinAsync(outer, inner, outerKeySelector, innerKeySelector, resultSelector) {
    internal_1.assert.mustBeAsyncQueryable(outer, "outer");
    internal_1.assert.mustBeAsyncQueryable(inner, "inner");
    internal_1.assert.mustBeFunction(outerKeySelector, "outerKeySelector");
    internal_1.assert.mustBeFunction(innerKeySelector, "innerKeySelector");
    internal_1.assert.mustBeFunction(resultSelector, "resultSelector");
    return new AsyncFullJoinIterable(internal_1.ToPossiblyAsyncIterable(outer), internal_1.ToPossiblyAsyncIterable(inner), outerKeySelector, innerKeySelector, resultSelector);
}
exports.fullJoinAsync = fullJoinAsync;
let AsyncFullJoinIterable = class AsyncFullJoinIterable {
    constructor(outer, inner, outerKeySelector, innerKeySelector, resultSelector) {
        this._outer = outer;
        this._inner = inner;
        this._outerKeySelector = outerKeySelector;
        this._innerKeySelector = innerKeySelector;
        this._resultSelector = resultSelector;
    }
    [Symbol.asyncIterator]() {
        return __asyncGenerator(this, arguments, function* _a() {
            const resultSelector = this._resultSelector;
            const outerLookup = new lookup_1.Lookup(yield __await(internal_1.CreateGroupingsAsync(this._outer, this._outerKeySelector, common_1.identity)));
            const innerLookup = new lookup_1.Lookup(yield __await(internal_1.CreateGroupingsAsync(this._inner, this._innerKeySelector, common_1.identity)));
            const keys = union_1.union(map_1.map(outerLookup, internal_1.SelectGroupingKey), map_1.map(innerLookup, internal_1.SelectGroupingKey));
            for (const key of keys) {
                const outer = defaultIfEmpty_1.defaultIfEmpty(outerLookup.get(key), undefined);
                const inner = defaultIfEmpty_1.defaultIfEmpty(innerLookup.get(key), undefined);
                for (const outerElement of outer) {
                    for (const innerElement of inner) {
                        yield yield __await(resultSelector(outerElement, innerElement));
                    }
                }
            }
        });
    }
};
AsyncFullJoinIterable = __decorate([
    internal_1.ToStringTag("AsyncFullJoinIterable")
], AsyncFullJoinIterable);
