"use strict";
/*!
  Copyright 2018 Ron Buckton (rbuckton@chronicles.org)

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
 */
/** @module "iterable-query/fn" */
var __values = (this && this.__values) || function (o) {
    var m = typeof Symbol === "function" && o[Symbol.iterator], i = 0;
    if (m) return m.call(o);
    return {
        next: function () {
            if (o && i >= o.length) o = void 0;
            return { value: o && o[i++], done: !o };
        }
    };
};
Object.defineProperty(exports, "__esModule", { value: true });
var internal_1 = require("../internal");
var common_1 = require("./common");
/**
 * Gets the minimum element of a [[Queryable]], optionally comparing elements using the supplied callback.
 *
 * @param source A [[Queryable]] object.
 * @param comparison An optional callback used to compare two elements.
 * @category Scalar
 */
function min(source, comparison) {
    if (comparison === void 0) { comparison = common_1.compare; }
    var e_1, _a;
    internal_1.assert.mustBeQueryable(source, "source");
    internal_1.assert.mustBeFunction(comparison, "comparison");
    var hasResult = false;
    var result;
    try {
        for (var _b = __values(internal_1.ToIterable(source)), _c = _b.next(); !_c.done; _c = _b.next()) {
            var element = _c.value;
            if (!hasResult) {
                result = element;
                hasResult = true;
            }
            else if (comparison(element, result) < 0) {
                result = element;
            }
        }
    }
    catch (e_1_1) { e_1 = { error: e_1_1 }; }
    finally {
        try {
            if (_c && !_c.done && (_a = _b.return)) _a.call(_b);
        }
        finally { if (e_1) throw e_1.error; }
    }
    return result;
}
exports.min = min;
