"use strict";
/*!
  Copyright 2018 Ron Buckton (rbuckton@chronicles.org)

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
 */
/** @module "iterable-query/fn" */
var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
var __generator = (this && this.__generator) || function (thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [op[0] & 2, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
};
var __values = (this && this.__values) || function (o) {
    var m = typeof Symbol === "function" && o[Symbol.iterator], i = 0;
    if (m) return m.call(o);
    return {
        next: function () {
            if (o && i >= o.length) o = void 0;
            return { value: o && o[i++], done: !o };
        }
    };
};
Object.defineProperty(exports, "__esModule", { value: true });
var internal_1 = require("../internal");
var types_1 = require("../types");
var toArray_1 = require("./toArray");
var common_1 = require("./common");
function orderBy(source, keySelector, comparison) {
    if (comparison === void 0) { comparison = common_1.compare; }
    internal_1.assert.mustBeQueryable(source, "source");
    internal_1.assert.mustBeFunction(keySelector, "keySelector");
    internal_1.assert.mustBeFunction(comparison, "comparison");
    return internal_1.FlowHierarchy(new OrderByIterable(internal_1.ToIterable(source), keySelector, comparison, /*descending*/ false), source);
}
exports.orderBy = orderBy;
function orderByDescending(source, keySelector, comparison) {
    if (comparison === void 0) { comparison = common_1.compare; }
    internal_1.assert.mustBeQueryable(source, "source");
    internal_1.assert.mustBeFunction(keySelector, "keySelector");
    internal_1.assert.mustBeFunction(comparison, "comparison");
    return internal_1.FlowHierarchy(new OrderByIterable(internal_1.ToIterable(source), keySelector, comparison, /*descending*/ true), source);
}
exports.orderByDescending = orderByDescending;
function thenBy(source, keySelector, comparison) {
    if (comparison === void 0) { comparison = common_1.compare; }
    internal_1.assert.mustBeOrderedIterable(source, "source");
    internal_1.assert.mustBeFunction(keySelector, "keySelector");
    internal_1.assert.mustBeFunction(comparison, "comparison");
    return internal_1.FlowHierarchy(internal_1.ThenBy(source, keySelector, comparison, /*descending*/ false), source);
}
exports.thenBy = thenBy;
function thenByDescending(source, keySelector, comparison) {
    if (comparison === void 0) { comparison = common_1.compare; }
    internal_1.assert.mustBeOrderedIterable(source, "source");
    internal_1.assert.mustBeFunction(keySelector, "keySelector");
    internal_1.assert.mustBeFunction(comparison, "comparison");
    return internal_1.FlowHierarchy(internal_1.ThenBy(source, keySelector, comparison, /*descending*/ true), source);
}
exports.thenByDescending = thenByDescending;
var OrderByIterable = /** @class */ (function () {
    function OrderByIterable(source, keySelector, comparison, descending, parent) {
        this._source = source;
        this._keySelector = keySelector;
        this._comparison = comparison;
        this._descending = descending;
        this._parent = parent;
    }
    OrderByIterable_1 = OrderByIterable;
    OrderByIterable.prototype[Symbol.iterator] = function () {
        var e_1, _a, source, array, sorter, len, indices, i, indices_1, indices_1_1, index, e_1_1;
        return __generator(this, function (_b) {
            switch (_b.label) {
                case 0:
                    source = this._source;
                    array = toArray_1.toArray(source);
                    sorter = this._getSorter(array);
                    len = array.length;
                    indices = new Array(len);
                    for (i = 0; i < len; ++i) {
                        indices[i] = i;
                    }
                    indices.sort(sorter);
                    _b.label = 1;
                case 1:
                    _b.trys.push([1, 6, 7, 8]);
                    indices_1 = __values(indices), indices_1_1 = indices_1.next();
                    _b.label = 2;
                case 2:
                    if (!!indices_1_1.done) return [3 /*break*/, 5];
                    index = indices_1_1.value;
                    return [4 /*yield*/, array[index]];
                case 3:
                    _b.sent();
                    _b.label = 4;
                case 4:
                    indices_1_1 = indices_1.next();
                    return [3 /*break*/, 2];
                case 5: return [3 /*break*/, 8];
                case 6:
                    e_1_1 = _b.sent();
                    e_1 = { error: e_1_1 };
                    return [3 /*break*/, 8];
                case 7:
                    try {
                        if (indices_1_1 && !indices_1_1.done && (_a = indices_1.return)) _a.call(indices_1);
                    }
                    finally { if (e_1) throw e_1.error; }
                    return [7 /*endfinally*/];
                case 8: return [2 /*return*/];
            }
        });
    };
    OrderByIterable.prototype[types_1.OrderedIterable.thenBy] = function (keySelector, comparison, descending) {
        internal_1.assert.mustBeFunction(keySelector, "keySelector");
        internal_1.assert.mustBeFunction(comparison, "comparison");
        internal_1.assert.mustBeBoolean(descending, "descending");
        return new OrderByIterable_1(this._source, keySelector, comparison, descending, this);
    };
    OrderByIterable.prototype._getSorter = function (elements, next) {
        var keySelector = this._keySelector;
        var comparison = this._comparison;
        var descending = this._descending;
        var parent = this._parent;
        var keys = elements.map(keySelector);
        var sorter = function (x, y) {
            var result = comparison(keys[x], keys[y]);
            if (result === 0) {
                return next ? next(x, y) : x - y;
            }
            return descending ? -result : result;
        };
        return parent ? parent._getSorter(elements, sorter) : sorter;
    };
    var OrderByIterable_1;
    OrderByIterable = OrderByIterable_1 = __decorate([
        internal_1.ToStringTag("OrderByIterable")
    ], OrderByIterable);
    return OrderByIterable;
}());
