"use strict";
/*!
  Copyright 2018 Ron Buckton (rbuckton@chronicles.org)

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
 */
/** @module "iterable-query/fn" */
var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
var __generator = (this && this.__generator) || function (thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [op[0] & 2, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
};
Object.defineProperty(exports, "__esModule", { value: true });
var internal_1 = require("../internal");
/**
 * Creates an [[Iterable]] over a range of numbers.
 *
 * @param start The starting number of the range.
 * @param end The ending number of the range.
 * @param increment The amount by which to change between each itereated value.
 * @category Query
 */
function range(start, end, increment) {
    if (increment === void 0) { increment = 1; }
    internal_1.assert.mustBeFiniteNumber(start, "start");
    internal_1.assert.mustBeFiniteNumber(end, "end");
    internal_1.assert.mustBePositiveNonZeroFiniteNumber(increment, "increment");
    return new RangeIterable(start, end, increment);
}
exports.range = range;
var RangeIterable = /** @class */ (function () {
    function RangeIterable(start, end, increment) {
        this._start = start;
        this._end = end;
        this._increment = increment;
    }
    RangeIterable.prototype[Symbol.iterator] = function () {
        var start, end, increment, i, i;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    start = this._start;
                    end = this._end;
                    increment = this._increment;
                    if (!(start <= end)) return [3 /*break*/, 5];
                    i = start;
                    _a.label = 1;
                case 1:
                    if (!(i <= end)) return [3 /*break*/, 4];
                    return [4 /*yield*/, i];
                case 2:
                    _a.sent();
                    _a.label = 3;
                case 3:
                    i += increment;
                    return [3 /*break*/, 1];
                case 4: return [3 /*break*/, 9];
                case 5:
                    i = start;
                    _a.label = 6;
                case 6:
                    if (!(i >= end)) return [3 /*break*/, 9];
                    return [4 /*yield*/, i];
                case 7:
                    _a.sent();
                    _a.label = 8;
                case 8:
                    i -= increment;
                    return [3 /*break*/, 6];
                case 9: return [2 /*return*/];
            }
        });
    };
    RangeIterable = __decorate([
        internal_1.ToStringTag("RangeIterable")
    ], RangeIterable);
    return RangeIterable;
}());
