"use strict";
/*!
  Copyright 2018 Ron Buckton (rbuckton@chronicles.org)

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
 */
/** @module "iterable-query" */
var __generator = (this && this.__generator) || function (thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [op[0] & 2, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
};
var __values = (this && this.__values) || function (o) {
    var m = typeof Symbol === "function" && o[Symbol.iterator], i = 0;
    if (m) return m.call(o);
    return {
        next: function () {
            if (o && i >= o.length) o = void 0;
            return { value: o && o[i++], done: !o };
        }
    };
};
var __read = (this && this.__read) || function (o, n) {
    var m = typeof Symbol === "function" && o[Symbol.iterator];
    if (!m) return o;
    var i = m.call(o), r, ar = [], e;
    try {
        while ((n === void 0 || n-- > 0) && !(r = i.next()).done) ar.push(r.value);
    }
    catch (error) { e = { error: error }; }
    finally {
        try {
            if (r && !r.done && (m = i["return"])) m.call(i);
        }
        finally { if (e) throw e.error; }
    }
    return ar;
};
Object.defineProperty(exports, "__esModule", { value: true });
var internal_1 = require("./internal");
var fn_1 = require("./fn");
var collections_1 = require("./collections");
var Lookup = /** @class */ (function () {
    /**
     * Creates a new Lookup for the provided groups.
     *
     * @param entries A map containing the unique groups of values.
     */
    function Lookup(entries) {
        internal_1.assert.mustBeQueryable(entries, "entries");
        this._entries = new collections_1.Map(internal_1.ToIterable(entries));
        this._source = new LookupIterable(this._entries, internal_1.CreateGrouping);
    }
    Object.defineProperty(Lookup.prototype, "size", {
        /**
         * Gets the number of unique keys.
         */
        get: function () {
            return this._entries.size;
        },
        enumerable: true,
        configurable: true
    });
    /**
     *
     * @param key A key.
     * Gets a value indicating whether any group has the provided key.
     */
    Lookup.prototype.has = function (key) {
        return this._entries.has(key);
    };
    /**
     * Gets the group for the provided key.
     *
     * @param key A key.
     */
    Lookup.prototype.get = function (key) {
        return internal_1.ToIterable(this._entries.get(key) || fn_1.empty());
    };
    /**
     * Creates a `Query` that maps each group in the Lookup.
     *
     * @param selector A callback used to select results for each group.
     */
    Lookup.prototype.applyResultSelector = function (selector) {
        internal_1.assert.mustBeFunction(selector, "selector");
        return new LookupIterable(this._entries, selector);
    };
    Lookup.prototype[Symbol.iterator] = function () {
        return internal_1.GetIterator(this._source);
    };
    return Lookup;
}());
exports.Lookup = Lookup;
var LookupIterable = /** @class */ (function () {
    function LookupIterable(map, selector) {
        this._map = map;
        this._selector = selector;
    }
    LookupIterable.prototype[Symbol.iterator] = function () {
        var e_1, _a, map, selector, map_1, map_1_1, _b, key, values, e_1_1;
        return __generator(this, function (_c) {
            switch (_c.label) {
                case 0:
                    map = this._map;
                    selector = this._selector;
                    _c.label = 1;
                case 1:
                    _c.trys.push([1, 6, 7, 8]);
                    map_1 = __values(map), map_1_1 = map_1.next();
                    _c.label = 2;
                case 2:
                    if (!!map_1_1.done) return [3 /*break*/, 5];
                    _b = __read(map_1_1.value, 2), key = _b[0], values = _b[1];
                    return [4 /*yield*/, selector(key, values)];
                case 3:
                    _c.sent();
                    _c.label = 4;
                case 4:
                    map_1_1 = map_1.next();
                    return [3 /*break*/, 2];
                case 5: return [3 /*break*/, 8];
                case 6:
                    e_1_1 = _c.sent();
                    e_1 = { error: e_1_1 };
                    return [3 /*break*/, 8];
                case 7:
                    try {
                        if (map_1_1 && !map_1_1.done && (_a = map_1.return)) _a.call(map_1);
                    }
                    finally { if (e_1) throw e_1.error; }
                    return [7 /*endfinally*/];
                case 8: return [2 /*return*/];
            }
        });
    };
    return LookupIterable;
}());
