"use strict";
/*!
  Copyright 2018 Ron Buckton (rbuckton@chronicles.org)

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
 */
/** @module "iterable-query/fn" */
var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
var __await = (this && this.__await) || function (v) { return this instanceof __await ? (this.v = v, this) : new __await(v); }
var __asyncGenerator = (this && this.__asyncGenerator) || function (thisArg, _arguments, generator) {
    if (!Symbol.asyncIterator) throw new TypeError("Symbol.asyncIterator is not defined.");
    var g = generator.apply(thisArg, _arguments || []), i, q = [];
    return i = {}, verb("next"), verb("throw"), verb("return"), i[Symbol.asyncIterator] = function () { return this; }, i;
    function verb(n) { if (g[n]) i[n] = function (v) { return new Promise(function (a, b) { q.push([n, v, a, b]) > 1 || resume(n, v); }); }; }
    function resume(n, v) { try { step(g[n](v)); } catch (e) { settle(q[0][3], e); } }
    function step(r) { r.value instanceof __await ? Promise.resolve(r.value.v).then(fulfill, reject) : settle(q[0][2], r); }
    function fulfill(value) { resume("next", value); }
    function reject(value) { resume("throw", value); }
    function settle(f, v) { if (f(v), q.shift(), q.length) resume(q[0][0], q[0][1]); }
};
Object.defineProperty(exports, "__esModule", { value: true });
var AsyncOrderByIterable_1;
"use strict";
const internal_1 = require("../internal");
const types_1 = require("../types");
const toArrayAsync_1 = require("./toArrayAsync");
const common_1 = require("./common");
function orderByAsync(source, keySelector, comparison = common_1.compare) {
    internal_1.assert.mustBeAsyncQueryable(source, "source");
    internal_1.assert.mustBeFunction(keySelector, "keySelector");
    internal_1.assert.mustBeFunction(comparison, "comparison");
    return internal_1.FlowHierarchy(new AsyncOrderByIterable(internal_1.ToPossiblyAsyncIterable(source), keySelector, comparison, /*descending*/ false), source);
}
exports.orderByAsync = orderByAsync;
function orderByDescendingAsync(source, keySelector, comparison = common_1.compare) {
    internal_1.assert.mustBeAsyncQueryable(source, "source");
    internal_1.assert.mustBeFunction(keySelector, "keySelector");
    internal_1.assert.mustBeFunction(comparison, "comparison");
    return internal_1.FlowHierarchy(new AsyncOrderByIterable(internal_1.ToPossiblyAsyncIterable(source), keySelector, comparison, /*descending*/ true), source);
}
exports.orderByDescendingAsync = orderByDescendingAsync;
function thenByAsync(source, keySelector, comparison = common_1.compare) {
    internal_1.assert.mustBePossiblyAsyncOrderedIterable(source, "source");
    internal_1.assert.mustBeFunction(keySelector, "keySelector");
    internal_1.assert.mustBeFunction(comparison, "comparison");
    return internal_1.FlowHierarchy(internal_1.ThenByAsync(internal_1.ToAsyncOrderedIterable(source), keySelector, comparison, /*descending*/ false), source);
}
exports.thenByAsync = thenByAsync;
function thenByDescendingAsync(source, keySelector, comparison = common_1.compare) {
    internal_1.assert.mustBePossiblyAsyncOrderedIterable(source, "source");
    internal_1.assert.mustBeFunction(keySelector, "keySelector");
    internal_1.assert.mustBeFunction(comparison, "comparison");
    return internal_1.FlowHierarchy(internal_1.ThenByAsync(internal_1.ToAsyncOrderedIterable(source), keySelector, comparison, /*descending*/ true), source);
}
exports.thenByDescendingAsync = thenByDescendingAsync;
let AsyncOrderByIterable = AsyncOrderByIterable_1 = class AsyncOrderByIterable {
    constructor(source, keySelector, comparison, descending, parent) {
        this._source = source;
        this._keySelector = keySelector;
        this._comparison = comparison;
        this._descending = descending;
        this._parent = parent;
    }
    [Symbol.asyncIterator]() {
        return __asyncGenerator(this, arguments, function* _a() {
            const source = this._source;
            const array = yield __await(toArrayAsync_1.toArrayAsync(source));
            const sorter = this._getSorter(array);
            const len = array.length;
            const indices = new Array(len);
            for (let i = 0; i < len; ++i) {
                indices[i] = i;
            }
            indices.sort(sorter);
            for (const index of indices) {
                yield yield __await(array[index]);
            }
        });
    }
    [types_1.AsyncOrderedIterable.thenByAsync](keySelector, comparison, descending) {
        internal_1.assert.mustBeFunction(keySelector, "keySelector");
        internal_1.assert.mustBeFunction(comparison, "comparison");
        internal_1.assert.mustBeBoolean(descending, "descending");
        return new AsyncOrderByIterable_1(this._source, keySelector, comparison, descending, this);
    }
    _getSorter(elements, next) {
        const keySelector = this._keySelector;
        const comparison = this._comparison;
        const descending = this._descending;
        const parent = this._parent;
        const keys = elements.map(keySelector);
        const sorter = (x, y) => {
            const result = comparison(keys[x], keys[y]);
            if (result === 0) {
                return next ? next(x, y) : x - y;
            }
            return descending ? -result : result;
        };
        return parent ? parent._getSorter(elements, sorter) : sorter;
    }
};
AsyncOrderByIterable = AsyncOrderByIterable_1 = __decorate([
    internal_1.ToStringTag("AsyncOrderByIterable")
], AsyncOrderByIterable);
