"use strict";
/*!
  Copyright 2018 Ron Buckton (rbuckton@chronicles.org)

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
 */
/** @module "iterable-query/fn" */
var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
Object.defineProperty(exports, "__esModule", { value: true });
const internal_1 = require("../internal");
function patch(source, start, skipCount = 0, range) {
    internal_1.assert.mustBeQueryable(source, "source");
    internal_1.assert.mustBePositiveFiniteNumber(start, "start");
    internal_1.assert.mustBePositiveFiniteNumber(skipCount, "skipCount");
    internal_1.assert.mustBeQueryableOrUndefined(range, "range");
    return internal_1.FlowHierarchy(new PatchIterable(internal_1.ToIterable(source), start, skipCount, range && internal_1.ToIterable(range)), source);
}
exports.patch = patch;
let PatchIterable = class PatchIterable {
    constructor(source, start, skipCount, range) {
        this._source = source;
        this._start = start;
        this._skipCount = skipCount;
        this._range = range;
    }
    *[Symbol.iterator]() {
        const start = this._start;
        const skipCount = this._skipCount;
        let offset = 0;
        let hasYieldedRange = false;
        for (const value of this._source) {
            if (offset < start) {
                yield value;
                offset++;
            }
            else if (offset < start + skipCount) {
                offset++;
            }
            else {
                if (!hasYieldedRange && this._range) {
                    yield* this._range;
                    hasYieldedRange = true;
                }
                yield value;
            }
        }
        if (!hasYieldedRange && this._range) {
            yield* this._range;
        }
    }
};
PatchIterable = __decorate([
    internal_1.ToStringTag("PatchIterable")
], PatchIterable);
