"use strict";
/*!
  Copyright 2018 Ron Buckton (rbuckton@chronicles.org)

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
 */
/** @module "iterable-query/fn" */
var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
var __await = (this && this.__await) || function (v) { return this instanceof __await ? (this.v = v, this) : new __await(v); }
var __asyncValues = (this && this.__asyncValues) || function (o) {
    if (!Symbol.asyncIterator) throw new TypeError("Symbol.asyncIterator is not defined.");
    var m = o[Symbol.asyncIterator], i;
    return m ? m.call(o) : (o = typeof __values === "function" ? __values(o) : o[Symbol.iterator](), i = {}, verb("next"), verb("throw"), verb("return"), i[Symbol.asyncIterator] = function () { return this; }, i);
    function verb(n) { i[n] = o[n] && function (v) { return new Promise(function (resolve, reject) { v = o[n](v), settle(resolve, reject, v.done, v.value); }); }; }
    function settle(resolve, reject, d, v) { Promise.resolve(v).then(function(v) { resolve({ value: v, done: d }); }, reject); }
};
var __asyncDelegator = (this && this.__asyncDelegator) || function (o) {
    var i, p;
    return i = {}, verb("next"), verb("throw", function (e) { throw e; }), verb("return"), i[Symbol.iterator] = function () { return this; }, i;
    function verb(n, f) { i[n] = o[n] ? function (v) { return (p = !p) ? { value: __await(o[n](v)), done: n === "return" } : f ? f(v) : v; } : f; }
};
var __asyncGenerator = (this && this.__asyncGenerator) || function (thisArg, _arguments, generator) {
    if (!Symbol.asyncIterator) throw new TypeError("Symbol.asyncIterator is not defined.");
    var g = generator.apply(thisArg, _arguments || []), i, q = [];
    return i = {}, verb("next"), verb("throw"), verb("return"), i[Symbol.asyncIterator] = function () { return this; }, i;
    function verb(n) { if (g[n]) i[n] = function (v) { return new Promise(function (a, b) { q.push([n, v, a, b]) > 1 || resume(n, v); }); }; }
    function resume(n, v) { try { step(g[n](v)); } catch (e) { settle(q[0][3], e); } }
    function step(r) { r.value instanceof __await ? Promise.resolve(r.value.v).then(fulfill, reject) : settle(q[0][2], r); }
    function fulfill(value) { resume("next", value); }
    function reject(value) { resume("throw", value); }
    function settle(f, v) { if (f(v), q.shift(), q.length) resume(q[0][0], q[0][1]); }
};
Object.defineProperty(exports, "__esModule", { value: true });
const internal_1 = require("../internal");
const types_1 = require("../types");
const axis_1 = require("./axis");
const collections_1 = require("../collections");
const toArrayAsync_1 = require("./toArrayAsync");
const common_1 = require("./common");
function topMostAsync(source, predicate = common_1.T) {
    internal_1.assert.mustBePossiblyAsyncHierarchyIterable(source, "source");
    return new AsyncTopMostIterable(source, predicate);
}
exports.topMostAsync = topMostAsync;
let AsyncTopMostIterable = class AsyncTopMostIterable {
    constructor(source, predicate) {
        this._source = source;
        this._predicate = predicate;
    }
    [Symbol.asyncIterator]() {
        return __asyncGenerator(this, arguments, function* _a() {
            const source = this._source;
            const predicate = this._predicate;
            const hierarchy = internal_1.GetHierarchy(source);
            const topMostNodes = yield __await(toArrayAsync_1.toArrayAsync(source));
            const ancestors = new collections_1.Map();
            for (let i = topMostNodes.length - 1; i >= 1; i--) {
                const node = topMostNodes[i];
                for (let j = i - 1; j >= 0; j--) {
                    const other = topMostNodes[j];
                    let ancestorsOfNode = ancestors.get(node);
                    if (!ancestorsOfNode) {
                        ancestorsOfNode = new collections_1.Set(axis_1.Axis.ancestors(hierarchy, node));
                        ancestors.set(node, ancestorsOfNode);
                    }
                    if (ancestorsOfNode.has(other)) {
                        topMostNodes.splice(i, 1);
                        break;
                    }
                    let ancestorsOfOther = ancestors.get(other);
                    if (!ancestorsOfOther) {
                        ancestorsOfOther = new collections_1.Set(axis_1.Axis.ancestors(hierarchy, other));
                        ancestors.set(other, ancestorsOfOther);
                    }
                    if (ancestorsOfOther.has(node)) {
                        topMostNodes.splice(j, 1);
                        i--;
                    }
                }
            }
            if (predicate === common_1.T) {
                yield __await(yield* __asyncDelegator(__asyncValues(topMostNodes)));
            }
            else {
                for (const node of topMostNodes) {
                    if (yield __await(predicate(node)))
                        yield yield __await(node);
                }
            }
        });
    }
    [types_1.Hierarchical.hierarchy]() {
        return internal_1.GetHierarchy(this._source);
    }
};
AsyncTopMostIterable = __decorate([
    internal_1.ToStringTag("AsyncTopMostIterable")
], AsyncTopMostIterable);
