"use strict";
/*!
  Copyright 2018 Ron Buckton (rbuckton@chronicles.org)

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
 */
/** @module "iterable-query/fn" */
var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
var __asyncValues = (this && this.__asyncValues) || function (o) {
    if (!Symbol.asyncIterator) throw new TypeError("Symbol.asyncIterator is not defined.");
    var m = o[Symbol.asyncIterator], i;
    return m ? m.call(o) : (o = typeof __values === "function" ? __values(o) : o[Symbol.iterator](), i = {}, verb("next"), verb("throw"), verb("return"), i[Symbol.asyncIterator] = function () { return this; }, i);
    function verb(n) { i[n] = o[n] && function (v) { return new Promise(function (resolve, reject) { v = o[n](v), settle(resolve, reject, v.done, v.value); }); }; }
    function settle(resolve, reject, d, v) { Promise.resolve(v).then(function(v) { resolve({ value: v, done: d }); }, reject); }
};
var __await = (this && this.__await) || function (v) { return this instanceof __await ? (this.v = v, this) : new __await(v); }
var __asyncGenerator = (this && this.__asyncGenerator) || function (thisArg, _arguments, generator) {
    if (!Symbol.asyncIterator) throw new TypeError("Symbol.asyncIterator is not defined.");
    var g = generator.apply(thisArg, _arguments || []), i, q = [];
    return i = {}, verb("next"), verb("throw"), verb("return"), i[Symbol.asyncIterator] = function () { return this; }, i;
    function verb(n) { if (g[n]) i[n] = function (v) { return new Promise(function (a, b) { q.push([n, v, a, b]) > 1 || resume(n, v); }); }; }
    function resume(n, v) { try { step(g[n](v)); } catch (e) { settle(q[0][3], e); } }
    function step(r) { r.value instanceof __await ? Promise.resolve(r.value.v).then(fulfill, reject) : settle(q[0][2], r); }
    function fulfill(value) { resume("next", value); }
    function reject(value) { resume("throw", value); }
    function settle(f, v) { if (f(v), q.shift(), q.length) resume(q[0][0], q[0][1]); }
};
Object.defineProperty(exports, "__esModule", { value: true });
const internal_1 = require("../internal");
const types_1 = require("../types");
const axis_1 = require("./axis");
const common_1 = require("./common");
function rootAsync(source, predicate = common_1.T) {
    internal_1.assert.mustBePossiblyAsyncHierarchyIterable(source, "source");
    internal_1.assert.mustBeFunctionOrUndefined(predicate, "predicate");
    return new AsyncRootTraversalIterable(source, predicate);
}
exports.rootAsync = rootAsync;
function ancestorsAsync(source, predicate = common_1.T) {
    internal_1.assert.mustBePossiblyAsyncHierarchyIterable(source, "source");
    internal_1.assert.mustBeFunctionOrUndefined(predicate, "predicate");
    return new AsyncAncestorsTraversalIterable(source, predicate);
}
exports.ancestorsAsync = ancestorsAsync;
function ancestorsAndSelfAsync(source, predicate = common_1.T) {
    internal_1.assert.mustBePossiblyAsyncHierarchyIterable(source, "source");
    internal_1.assert.mustBeFunctionOrUndefined(predicate, "predicate");
    return new AsyncAncestorsAndSelfTraversalIterable(source, predicate);
}
exports.ancestorsAndSelfAsync = ancestorsAndSelfAsync;
function descendantsAsync(source, predicate = common_1.T) {
    internal_1.assert.mustBePossiblyAsyncHierarchyIterable(source, "source");
    internal_1.assert.mustBeFunctionOrUndefined(predicate, "predicate");
    return new AsyncDescendantsTraversalIterable(source, predicate);
}
exports.descendantsAsync = descendantsAsync;
function descendantsAndSelfAsync(source, predicate = common_1.T) {
    internal_1.assert.mustBePossiblyAsyncHierarchyIterable(source, "source");
    internal_1.assert.mustBeFunctionOrUndefined(predicate, "predicate");
    return new AsyncDescendantsAndSelfTraversalIterable(source, predicate);
}
exports.descendantsAndSelfAsync = descendantsAndSelfAsync;
function parentsAsync(source, predicate = common_1.T) {
    internal_1.assert.mustBePossiblyAsyncHierarchyIterable(source, "source");
    internal_1.assert.mustBeFunctionOrUndefined(predicate, "predicate");
    return new AsyncParentsTraversalIterable(source, predicate);
}
exports.parentsAsync = parentsAsync;
function selfAsync(source, predicate = common_1.T) {
    internal_1.assert.mustBePossiblyAsyncHierarchyIterable(source, "source");
    internal_1.assert.mustBeFunctionOrUndefined(predicate, "predicate");
    return new AsyncSelfTraversalIterable(source, predicate);
}
exports.selfAsync = selfAsync;
function siblingsAsync(source, predicate = common_1.T) {
    internal_1.assert.mustBePossiblyAsyncHierarchyIterable(source, "source");
    internal_1.assert.mustBeFunctionOrUndefined(predicate, "predicate");
    return new AsyncSiblingsTraversalIterable(source, predicate);
}
exports.siblingsAsync = siblingsAsync;
function siblingsAndSelfAsync(source, predicate = common_1.T) {
    internal_1.assert.mustBePossiblyAsyncHierarchyIterable(source, "source");
    internal_1.assert.mustBeFunctionOrUndefined(predicate, "predicate");
    return new AsyncSiblingsAndSelfTraversalIterable(source, predicate);
}
exports.siblingsAndSelfAsync = siblingsAndSelfAsync;
function siblingsBeforeSelfAsync(source, predicate = common_1.T) {
    internal_1.assert.mustBePossiblyAsyncHierarchyIterable(source, "source");
    internal_1.assert.mustBeFunctionOrUndefined(predicate, "predicate");
    return new AsyncSiblingsBeforeSelfTraversalIterable(source, predicate);
}
exports.siblingsBeforeSelfAsync = siblingsBeforeSelfAsync;
function siblingsAfterSelfAsync(source, predicate = common_1.T) {
    internal_1.assert.mustBePossiblyAsyncHierarchyIterable(source, "source");
    internal_1.assert.mustBeFunctionOrUndefined(predicate, "predicate");
    return new AsyncSiblingsAfterSelfTraversalIterable(source, predicate);
}
exports.siblingsAfterSelfAsync = siblingsAfterSelfAsync;
function childrenAsync(source, predicate = common_1.T) {
    internal_1.assert.mustBePossiblyAsyncHierarchyIterable(source, "source");
    internal_1.assert.mustBeFunctionOrUndefined(predicate, "predicate");
    return new AsyncChildrenTraversalIterable(source, predicate);
}
exports.childrenAsync = childrenAsync;
function CreateAsyncTraversalIterable(tag, axis) {
    let AsyncHierarchyTraversalIterable = class AsyncHierarchyTraversalIterable {
        constructor(source, predicate) {
            this._source = source;
            this._predicate = predicate;
            this._axis = axis;
        }
        [Symbol.asyncIterator]() {
            return __asyncGenerator(this, arguments, function* _a() {
                var e_1, _a;
                const source = this._source;
                const hierarchy = internal_1.GetHierarchy(this._source);
                const predicate = this._predicate;
                const axis = this._axis;
                try {
                    for (var source_1 = __asyncValues(source), source_1_1; source_1_1 = yield __await(source_1.next()), !source_1_1.done;) {
                        const element = source_1_1.value;
                        for (const related of axis(hierarchy, element)) {
                            if (yield __await(predicate(related))) {
                                yield yield __await(related);
                            }
                        }
                    }
                }
                catch (e_1_1) { e_1 = { error: e_1_1 }; }
                finally {
                    try {
                        if (source_1_1 && !source_1_1.done && (_a = source_1.return)) yield __await(_a.call(source_1));
                    }
                    finally { if (e_1) throw e_1.error; }
                }
            });
        }
        [types_1.Hierarchical.hierarchy]() {
            return internal_1.GetHierarchy(this._source);
        }
    };
    AsyncHierarchyTraversalIterable = __decorate([
        internal_1.ClassName(tag),
        internal_1.ToStringTag(tag)
    ], AsyncHierarchyTraversalIterable);
    return AsyncHierarchyTraversalIterable;
}
const AsyncRootTraversalIterable = CreateAsyncTraversalIterable("AsyncRootTraversalIterable", axis_1.Axis.root);
const AsyncAncestorsTraversalIterable = CreateAsyncTraversalIterable("AsyncAncestorsTraversalIterable", axis_1.Axis.ancestors);
const AsyncAncestorsAndSelfTraversalIterable = CreateAsyncTraversalIterable("AsyncAncestorsAndSelfTraversalIterable", axis_1.Axis.ancestorsAndSelf);
const AsyncDescendantsTraversalIterable = CreateAsyncTraversalIterable("AsyncDescendantsTraversalIterable", axis_1.Axis.descendants);
const AsyncDescendantsAndSelfTraversalIterable = CreateAsyncTraversalIterable("AsyncDescendantsAndSelfTraversalIterable", axis_1.Axis.descendantsAndSelf);
const AsyncParentsTraversalIterable = CreateAsyncTraversalIterable("AsyncParentsTraversalIterable", axis_1.Axis.parents);
const AsyncSelfTraversalIterable = CreateAsyncTraversalIterable("AsyncSelfTraversalIterable", axis_1.Axis.self);
const AsyncSiblingsTraversalIterable = CreateAsyncTraversalIterable("AsyncSiblingsTraversalIterable", axis_1.Axis.siblings);
const AsyncSiblingsAndSelfTraversalIterable = CreateAsyncTraversalIterable("AsyncSiblingsAndSelfTraversalIterable", axis_1.Axis.siblingsAndSelf);
const AsyncSiblingsBeforeSelfTraversalIterable = CreateAsyncTraversalIterable("AsyncSiblingsBeforeSelfTraversalIterable", axis_1.Axis.siblingsBeforeSelf);
const AsyncSiblingsAfterSelfTraversalIterable = CreateAsyncTraversalIterable("AsyncSiblingsAfterSelfTraversalIterable", axis_1.Axis.siblingsAfterSelf);
const AsyncChildrenTraversalIterable = CreateAsyncTraversalIterable("AsyncChildrenTraversalIterable", axis_1.Axis.children);
