"use strict";
/*!
  Copyright 2018 Ron Buckton (rbuckton@chronicles.org)

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
 */
/** @module "iterable-query/fn" */
var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
Object.defineProperty(exports, "__esModule", { value: true });
const internal_1 = require("../internal");
const common_1 = require("./common");
function zip(left, right, selector = common_1.tuple) {
    internal_1.assert.mustBeQueryable(left, "left");
    internal_1.assert.mustBeQueryable(right, "right");
    internal_1.assert.mustBeFunction(selector, "selector");
    return new ZipIterable(internal_1.ToIterable(left), internal_1.ToIterable(right), selector);
}
exports.zip = zip;
let ZipIterable = class ZipIterable {
    constructor(left, right, selector) {
        this._left = left;
        this._right = right;
        this._selector = selector;
    }
    *[Symbol.iterator]() {
        const selector = this._selector;
        const leftIterator = internal_1.GetIterator(this._left);
        let leftDone = false;
        let leftValue;
        try {
            const rightIterator = internal_1.GetIterator(this._right);
            let rightDone = false;
            let rightValue;
            try {
                for (;;) {
                    ({ done: leftDone, value: leftValue } = leftIterator.next());
                    ({ done: rightDone, value: rightValue } = rightIterator.next());
                    if (leftDone || rightDone)
                        break;
                    yield selector(leftValue, rightValue);
                }
            }
            finally {
                if (!rightDone)
                    internal_1.IteratorClose(rightIterator);
            }
        }
        finally {
            if (!leftDone)
                internal_1.IteratorClose(leftIterator);
        }
    }
};
ZipIterable = __decorate([
    internal_1.ToStringTag("ZipIterable")
], ZipIterable);
